from .utils import read_json_list, evaluate_function_call, extract_function_call_json
import copy
import json
import random
import string
from .graphql_tool import OpentargetTool, OpentargetGeneticsTool, OpentargetToolDrugNameMatch
from .openfda_tool import FDADrugLabelTool, FDADrugLabelSearchTool, FDADrugLabelSearchIDTool, FDADrugLabelGetDrugGenericNameTool
from .restful_tool import MonarchTool, MonarchDiseasesForMultiplePhenoTool

import os

# Determine the directory where the current file is located
current_dir = os.path.dirname(os.path.abspath(__file__))

default_tool_files = {
    'opentarget': os.path.join(current_dir, 'data', 'opentarget_tools.json'),
    'fda_drug_label': os.path.join(current_dir, 'data', 'fda_drug_labeling_tools.json'),
    'special_tools': os.path.join(current_dir, 'data', 'special_tools.json'),
    'monarch': os.path.join(current_dir, 'data', 'monarch_tools.json')
}

tool_type_mappings = {
    'OpenTarget': OpentargetTool,
    'OpenTargetGenetics': OpentargetGeneticsTool,
    'FDADrugLabel': FDADrugLabelTool,
    'FDADrugLabelSearchTool': FDADrugLabelSearchTool,
    'Monarch': MonarchTool,
    'MonarchDiseasesForMultiplePheno': MonarchDiseasesForMultiplePhenoTool,
    'FDADrugLabelSearchIDTool': FDADrugLabelSearchIDTool,
    'FDADrugLabelGetDrugGenericNameTool': FDADrugLabelGetDrugGenericNameTool,
    'OpentargetToolDrugNameMatch': OpentargetToolDrugNameMatch,
}


class ToolUniverse:
    def __init__(self, tool_files=default_tool_files, keep_default_tools=True):
        # Initialize any necessary attributes here
        self.all_tools = []
        self.all_tool_dict = {}
        self.tool_category_dicts = {}
        if tool_files is None:
            tool_files = default_tool_files
        elif keep_default_tools:
            default_tool_files.update(tool_files)
            tool_files = default_tool_files
        self.tool_files = tool_files
        print("Tool files:")
        print(tool_files)
        self.callable_functions = {}

    def load_tools(self, tool_type=None):
        print(f"Number of tools before load tools: {len(self.all_tools)}")
        if tool_type is None:
            for each in self.tool_files:
                loaded_tool_list = read_json_list(self.tool_files[each])
                self.all_tools += loaded_tool_list
                self.tool_category_dicts[each] = loaded_tool_list
        else:
            for each in tool_type:
                loaded_tool_list = read_json_list(self.tool_files[each])
                self.all_tools += loaded_tool_list
                self.tool_category_dicts[each] = loaded_tool_list
        # Deduplication of tools
        tool_name_list = []
        dedup_all_tools = []
        for each in self.all_tools:
            if each['name'] not in tool_name_list:
                tool_name_list.append(each['name'])
                dedup_all_tools.append(each)
        self.all_tools = dedup_all_tools
        self.refresh_tool_name_desc()

        print(f"Number of tools after load tools: {len(self.all_tools)}")

    def return_all_loaded_tools(self):
        return copy.deepcopy(self.all_tools)

    def refresh_tool_name_desc(self, enable_full_desc=False):
        tool_name_list = []
        tool_desc_list = []
        for tool in self.all_tools:
            tool_name_list.append(tool['name'])
            if enable_full_desc:
                tool_desc_list.append(json.dumps(tool))
            else:
                tool_desc_list.append(tool['name']+': '+tool['description'])
            self.all_tool_dict[tool['name']] = tool
        return tool_name_list, tool_desc_list

    def prepare_one_tool_prompt(self, tool):
        valid_keys = ['name', 'description', 'parameter', 'required']
        tool = copy.deepcopy(tool)
        for key in list(tool.keys()):
            if key not in valid_keys:
                del tool[key]
        return tool
        
    def prepare_tool_prompts(self, tool_list):
        copied_list = []
        for tool in tool_list:
            copied_list.append(self.prepare_one_tool_prompt(tool))
        return copied_list

    def remove_keys(self, tool_list, invalid_keys):
        copied_list = copy.deepcopy(tool_list)
        for tool in copied_list:
            # Create a list of keys to avoid modifying the dictionary during iteration
            for key in list(tool.keys()):
                if key in invalid_keys:
                    del tool[key]
        return copied_list

    def prepare_tool_examples(self, tool_list):
        valid_keys = ['name', 'description',
                      'parameter', 'required', 'query_schema', 'fields', 'label', 'type']
        copied_list = copy.deepcopy(tool_list)
        for tool in copied_list:
            # Create a list of keys to avoid modifying the dictionary during iteration
            for key in list(tool.keys()):
                if key not in valid_keys:
                    del tool[key]
        return copied_list

    def get_tool_by_name(self, tool_names):
        picked_tool_list = []
        for each_name in tool_names:
            if each_name in self.all_tool_dict:
                picked_tool_list.append(self.all_tool_dict[each_name])
            else:
                print(f"Tool name {each_name} not found in the loaded tools.")
        return picked_tool_list

    def get_one_tool_by_one_name(self, tool_name, return_prompt=False):
            if tool_name in self.all_tool_dict:
                if return_prompt:
                    return self.prepare_one_tool_prompt(self.all_tool_dict[tool_name])
                return self.all_tool_dict[tool_name]
            else:
                print(f"Tool name {tool_name} not found in the loaded tools.")
                return None

    def get_tool_type_by_name(self, tool_name):
        return self.all_tool_dict[tool_name]['type']

    def tool_to_str(self, tool_list):
        return '\n\n'.join(json.dumps(obj, indent=4) for obj in tool_list)

    def extract_function_call_json(self, lst, return_message=False, verbose=True):
        return extract_function_call_json(lst, return_message=return_message, verbose=verbose)

    def call_id_gen(self):
        return "".join(random.choices(string.ascii_letters + string.digits, k=9))

    def run(self, fcall_str, return_message=False, verbose=True):
        if return_message:
            function_call_json, message = self.extract_function_call_json(
                fcall_str, return_message=return_message, verbose=verbose)
        else:
            function_call_json = self.extract_function_call_json(
                fcall_str, return_message=return_message, verbose=verbose)
        if function_call_json is not None:
            if isinstance(function_call_json, list):
                # return the function call+result message with call id.
                call_results = []
                for i in range(len(function_call_json)):
                    call_result = self.run_one_function(function_call_json[i])
                    call_id = self.call_id_gen()
                    function_call_json[i]["call_id"] = call_id
                    call_results.append({"role": "tool", "content": json.dumps(
                        {"content": call_result, "call_id": call_id})})
                revised_messages = [{"role": "assistant", "content": message,
                                     "tool_calls": json.dumps(function_call_json)}]+call_results
                return revised_messages
            else:
                return self.run_one_function(function_call_json)
        else:
            print("\033[91mNot a function call\033[0m")
            return None

    def run_one_function(self, function_call_json):
        check_status, check_message = self.check_function_call(function_call_json)
        if check_status is False:
            return "Invalid function call: "+check_message# + "  You must correct your invalid function call!"
        function_name = function_call_json["name"]
        arguments = function_call_json["arguments"]
        if function_name in self.callable_functions:
            return self.callable_functions[function_name].run(arguments)
        else:
            if function_name in self.all_tool_dict:
                print(
                    "\033[92mInitiating callable_function from loaded tool dicts.\033[0m")
                tool = self.init_tool(
                    self.all_tool_dict[function_name], add_to_cache=True)
                return tool.run(arguments)

    def init_tool(self, tool=None, tool_name=None, add_to_cache=True):
        if tool_name is not None:
            new_tool = tool_type_mappings[tool_name]()
        else:
            tool_type = tool['type']
            tool_name = tool['name']
            if 'OpentargetToolDrugNameMatch' == tool_type:
                if 'FDADrugLabelGetDrugGenericNameTool' not in self.callable_functions:
                    self.callable_functions['FDADrugLabelGetDrugGenericNameTool'] = tool_type_mappings['FDADrugLabelGetDrugGenericNameTool']()
                new_tool = tool_type_mappings[tool_type](tool_config=tool, drug_generic_tool=self.callable_functions['FDADrugLabelGetDrugGenericNameTool'])
            else:
                new_tool = tool_type_mappings[tool_type](tool_config=tool)
        if add_to_cache:
            self.callable_functions[tool_name] = new_tool
        return new_tool

    def check_function_call(self, fcall_str, function_config=None):
        function_call_json = self.extract_function_call_json(fcall_str)
        print("loaded function call json", function_call_json)
        if function_call_json is not None:
            if function_config is not None:
                return evaluate_function_call(function_config, function_call_json)
            function_name = function_call_json['name']
            if not  function_name in self.all_tool_dict:
                return False, f"Function name {function_name} not found in loaded tools."
            return evaluate_function_call(self.all_tool_dict[function_name], function_call_json)
        else:
            return False, "\033[91mInvalid JSON string of function call\033[0m"



import yaml
import json


def yaml_to_dict(yaml_file_path):
    """
    Convert a YAML file to a dictionary.

    Args:
        yaml_file_path (str): Path to the YAML file.

    Returns:
        dict: Dictionary representation of the YAML file content.
    """
    try:
        with open(yaml_file_path, 'r') as file:
            yaml_dict = yaml.safe_load(file)
            return yaml_dict
    except FileNotFoundError:
        print(f"File not found: {yaml_file_path}")
    except yaml.YAMLError as exc:
        print(f"Error in YAML file: {exc}")


def read_json_list(file_path):
    """
    Reads a list of JSON objects from a file.

    Parameters:
    file_path (str): The path to the JSON file.

    Returns:
    list: A list of dictionaries containing the JSON objects.
    """
    with open(file_path, 'r') as file:
        data = json.load(file)
    return data


def evaluate_function_call(tool_definition, function_call):
    # Map for type conversion
    type_map = {
        "string": str,
        "integer": int,
        "number": float,
        "boolean": bool,
        "array": list,
        "object": dict
    }

    # Check if the function name matches
    if tool_definition["name"] != function_call["name"]:
        return False, "Function name does not match."

    # Check if all required parameters are present
    required_params = [key for key, value in tool_definition["parameter"]
                       ["properties"].items() if value.get("required", False)]
    missing_params = [
        param for param in required_params if param not in function_call["arguments"]]
    if missing_params:
        return False, f"Missing required parameters: {missing_params}"

    # Check if all provided parameters are valid and their data types are correct
    valid_params = tool_definition["parameter"]["properties"]
    invalid_params = []
    type_mismatches = []

    for param, value in function_call["arguments"].items():
        if param not in valid_params:
            invalid_params.append(param)
        else:
            expected_type = valid_params[param]["type"]
            if expected_type not in type_map:
                return False, f"Unsupported parameter type: {expected_type}"
            if not isinstance(value, type_map[expected_type]):
                type_mismatches.append(
                    (param, expected_type, type(value).__name__))

    if invalid_params:
        return False, f"Invalid parameters provided: {invalid_params}"

    if type_mismatches:
        return False, f"Type mismatches: {type_mismatches}"

    return True, "Function call is valid."

def evaluate_function_call_from_toolbox(toolbox, function_call):
    tool_name = function_call["name"]
    this_tool_dec = toolbox.get_one_tool_by_one_name(tool_name)
    if this_tool_dec is None:
        return False, "Tool not found."
    results, results_message = evaluate_function_call(this_tool_dec, function_call)
    return results, results_message
    

def compare_function_calls(pred_function_call, gt_function_call, compare_arguments=True, compare_value=True):
    # Extracting the name and arguments from the predicted function call
    pred_name = pred_function_call["name"]
    pred_arguments = pred_function_call["arguments"]

    # Extracting the name and arguments from the ground truth function call
    gt_name = gt_function_call["name"]
    gt_arguments = gt_function_call["arguments"]

    # Compare function names
    if pred_name != gt_name:
        return False, "Function names do not match."

    if compare_arguments:
        # Compare arguments
        if set(pred_arguments.keys()) != set(gt_arguments.keys()):
            missing_in_pred = set(gt_arguments.keys()) - set(pred_arguments.keys())
            missing_in_gt = set(pred_arguments.keys()) - set(gt_arguments.keys())
            return False, f"Argument keys do not match. Missing in predicted: {missing_in_pred}, Missing in ground truth: {missing_in_gt}"
    if compare_value:
        # Compare argument values
        mismatched_values = []
        for key in pred_arguments:
            if pred_arguments[key] != gt_arguments[key]:
                mismatched_values.append((key, pred_arguments[key], gt_arguments[key]))

        if mismatched_values:
            return False, f"Argument values do not match: {mismatched_values}"

    return True, "Function calls match."


def extract_function_call_json(lst, return_message=False, verbose=True):
    if type(lst) is dict:
        if return_message:
            return lst, ""
        return lst
    result_str = ''.join(lst)
    if verbose:
        print("\033[1;34mPossible LLM outputs for function call:\033[0m", result_str)
    try:
        function_call_json = json.loads(result_str.strip())
        if return_message:
            return function_call_json, ""
        return function_call_json
    except json.JSONDecodeError:
        try:
            index_start = result_str.find(
                '[TOOL_CALLS]') 
            index_end = result_str.find('</s>')
            if index_end == -1:
                index_end = result_str.find('<|eom_id|>')
            if index_end == -1:
                function_call_str = result_str[index_start+ len('[TOOL_CALLS]'):]
            else:
                function_call_str = result_str[index_start+ len('[TOOL_CALLS]'):index_end]
            # print("function_call_str", function_call_str)
            function_call_json = json.loads(function_call_str.strip())
            if return_message:
                message = result_str[:index_start]
                return function_call_json, message
            return function_call_json
        except json.JSONDecodeError:
            try:
                print("Multiple function calls not implemented for 'llama' format.")
                index_start = result_str.find(
                    '<functioncall>') + len('<functioncall>')
                index_end = result_str.find('</functioncall>')
                function_call_str = result_str[index_start:index_end]
                # function_call_str = function_call_str.replace("'", '"')
                function_call_json = json.loads(function_call_str.strip())
                return function_call_json
            except json.JSONDecodeError as e:
                print("Not a function call:", e)
                if return_message:
                    return None, result_str
                return None



from graphql import build_schema
from graphql.language import parse
from graphql.validation import validate
from .base_tool import BaseTool
import requests
import copy
import json


def validate_query(query_str, schema_str):
    try:
        # Build the GraphQL schema object from the provided schema string
        schema = build_schema(schema_str)

        # Parse the query string into an AST (Abstract Syntax Tree)
        query_ast = parse(query_str)

        # Validate the query AST against the schema
        validation_errors = validate(schema, query_ast)

        if not validation_errors:
            return True
        else:
            # Collect and return the validation errors
            error_messages = '\n'.join(str(error)
                                       for error in validation_errors)
            return f"Query validation errors:\n{error_messages}"
    except Exception as e:
        return f"An error occurred during validation: {str(e)}"

def remove_none_and_empty_values(json_obj):
    """Remove all key-value pairs where the value is None or an empty list"""
    if isinstance(json_obj, dict):
        return {k: remove_none_and_empty_values(v) for k, v in json_obj.items() if v is not None and v != []}
    elif isinstance(json_obj, list):
        return [remove_none_and_empty_values(item) for item in json_obj if item is not None and item != []]
    else:
        return json_obj

def execute_query(endpoint_url, query, variables=None):
    response = requests.post(
        endpoint_url, json={'query': query, 'variables': variables})
    try:
        result = response.json()
        # result = json.dumps(result, ensure_ascii=False)
        result = remove_none_and_empty_values(result)
        # Check if the response contains errors
        if 'errors' in result:
            print("Invalid Query: ", result['errors'])
            return None
         # Check if the data field is empty
        elif not result.get('data') or all(not v for v in result['data'].values()):
            print("No data returned")
            return None
        else:
            return result
    except requests.exceptions.JSONDecodeError as e:
        print("JSONDecodeError: Could not decode the response as JSON")
        return None


class GraphQLTool(BaseTool):
    def __init__(self, tool_config, endpoint_url):
        super().__init__(tool_config)
        self.endpoint_url = endpoint_url
        self.query_schema = tool_config['query_schema']
        self.parameters = tool_config['parameter']['properties']
        self.default_size = 5

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        if 'size' in self.parameters and 'size' not in arguments:
            arguments['size'] = 5
        return execute_query(endpoint_url=self.endpoint_url, query=self.query_schema, variables=arguments)


class OpentargetTool(GraphQLTool):
    def __init__(self, tool_config):
        endpoint_url = 'https://api.platform.opentargets.org/api/v4/graphql'
        super().__init__(tool_config, endpoint_url)
        
    def run(self, arguments):
        for each_arg, arg_value in arguments.items(): # opentarget api cannot handle '-' in the arguments
            if isinstance(arg_value, str): 
                if '-' in arg_value:
                    arguments[each_arg] = arg_value.replace('-', ' ')
        return super().run(arguments)
            

class OpentargetToolDrugNameMatch(GraphQLTool):
    def __init__(self, tool_config, drug_generic_tool=None):
        endpoint_url = 'https://api.platform.opentargets.org/api/v4/graphql'
        self.drug_generic_tool = drug_generic_tool
        self.possible_drug_name_args = ['drugName']
        super().__init__(tool_config, endpoint_url)

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        results = execute_query(endpoint_url=self.endpoint_url, query=self.query_schema, variables=arguments)
        if results is None:
            print("No results found for the drug brand name. Trying with the generic name.")
            name_arguments = {}
            for each_args in self.possible_drug_name_args:
                if each_args in arguments:
                    name_arguments['drug_name'] = arguments[each_args]
                    break
            if len(name_arguments)==0:
                print("No drug name found in the arguments.")
                return None
            drug_name_results = self.drug_generic_tool.run(name_arguments)
            if drug_name_results is not None and 'openfda.generic_name' in drug_name_results:
                arguments[each_args] = drug_name_results['openfda.generic_name']
                print("Found generic name. Trying with the generic name: ", arguments[each_args])
                results = execute_query(endpoint_url=self.endpoint_url, query=self.query_schema, variables=arguments)
        return results
            


class OpentargetGeneticsTool(GraphQLTool):
    def __init__(self, tool_config):
        endpoint_url = 'https://api.genetics.opentargets.org/graphql'
        super().__init__(tool_config, endpoint_url)



from .graphql_tool import GraphQLTool
import requests
import copy
import json

def execute_RESTful_query(endpoint_url, variables=None):
    response = requests.get(
        endpoint_url, params = variables)
    try:
        result = response.json()

        # Check if the response contains errors
        if 'error' in result:
            print("Invalid Query: ", result['error'])
            return False
        else:
            return result
    except requests.exceptions.JSONDecodeError as e:
        print("JSONDecodeError: Could not decode the response as JSON")
        return False
    except requests.exceptions.HTTPError as e:
        print(f"HTTP error occurred: {e}")
    except Exception as e:
        print(f"An error occurred: {e}")

class RESTfulTool(GraphQLTool):
    def __init__(self, tool_config, endpoint_url):
        super().__init__(tool_config, endpoint_url)

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        return execute_RESTful_query(endpoint_url=self.endpoint_url, variables=arguments)

class MonarchTool(RESTfulTool):
    def __init__(self, tool_config):
        endpoint_url = 'https://api.monarchinitiative.org/v3/api' + tool_config['tool_url']
        super().__init__(tool_config, endpoint_url)
    
    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        query_schema_runtime = copy.deepcopy(self.query_schema)
        for key in query_schema_runtime:
            if key in arguments:
                query_schema_runtime[key] = arguments[key]
        if "url_key" in query_schema_runtime:
            url_key_name = query_schema_runtime['url_key']
            formatted_endpoint_url = self.endpoint_url.format(url_key = query_schema_runtime[url_key_name])
            del query_schema_runtime['url_key']
        else:
            formatted_endpoint_url = self.endpoint_url
        if isinstance(query_schema_runtime, dict):
            print(query_schema_runtime)
            if 'query' in query_schema_runtime:
                query_schema_runtime['q'] = query_schema_runtime['query'] # match with the api
        response = execute_RESTful_query(endpoint_url=formatted_endpoint_url, variables=query_schema_runtime)
        if 'facet_fields' in response:
            del response['facet_fields']
        def remove_empty_values(obj):
            if isinstance(obj, dict):
                return {k: remove_empty_values(v) for k, v in obj.items()
                        if v not in [0, [], None]}
            elif isinstance(obj, list):
                return [remove_empty_values(v) for v in obj if v not in [0, [], None]]
            else:
                return obj
        response = remove_empty_values(response)
        return response

class MonarchDiseasesForMultiplePhenoTool(MonarchTool):
    def __init__(self, tool_config):
        super().__init__(tool_config)
    
    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        query_schema_runtime = copy.deepcopy(self.query_schema)
        for key in query_schema_runtime:
            if (key!="HPO_ID_list") and (key in arguments):
                query_schema_runtime[key] = arguments[key]
        all_diseases = []
        for HPOID in arguments['HPO_ID_list']:
            each_query_schema_runtime = copy.deepcopy(query_schema_runtime)
            each_query_schema_runtime['object'] = HPOID
            each_query_schema_runtime['limit'] = 500
            each_output = execute_RESTful_query(endpoint_url=self.endpoint_url, variables=each_query_schema_runtime)
            each_output = each_output['items']
            each_output_names = [disease['subject_label'] for disease in each_output]
            all_diseases.append(each_output_names)
        
        intersection = set(all_diseases[0])
        for element in all_diseases[1:]:
            intersection &= set(element)
        intersection = list(intersection)
        if query_schema_runtime['limit'] < len(intersection):
            intersection = intersection[:query_schema_runtime['limit']]
        return intersection



from .utils import extract_function_call_json, evaluate_function_call

class BaseTool:
    def __init__(self, tool_config):
        self.tool_config = tool_config

    def run(self):
        pass

    def check_function_call(self, function_call_json):
        if isinstance(function_call_json, str):
            function_call_json = extract_function_call_json(function_call_json)
        if function_call_json is not None:
            return evaluate_function_call(self.tool_config, function_call_json)
        else:
            return False, "Invalid JSON string of function call"

    def get_required_parameters(self):
        """
        Retrieve required parameters from the endpoint definition.
        Returns:
        list: List of required parameters for the given endpoint.
        """
        required_params = []
        parameters = self.tool_config.get('parameter', {}).get('properties', {})

        # Check each parameter to see if it is required
        for param, details in parameters.items():
            if details.get('required', False):
                required_params.append(param.lower())

        return required_params


from .execute_function import ToolUniverse
from .restful_tool import MonarchTool, MonarchDiseasesForMultiplePhenoTool
from .graphql_tool import OpentargetTool, OpentargetGeneticsTool, OpentargetToolDrugNameMatch
from .openfda_tool import FDADrugLabelTool, FDADrugLabelSearchTool, FDADrugLabelSearchIDTool, FDADrugLabelGetDrugGenericNameTool

__all__ = [
    "ToolUniverse",
    "MonarchTool",
    "MonarchDiseasesForMultiplePhenoTool",
    "OpentargetTool",
    "OpentargetGeneticsTool",
    "OpentargetToolDrugNameMatch",
    "FDADrugLabelTool",
    "FDADrugLabelSearchTool",
    "FDADrugLabelSearchIDTool",
    "FDADrugLabelGetDrugGenericNameTool"
]



import requests
from .base_tool import BaseTool
import copy
import re
import json
import os

def check_keys_present(api_capabilities_dict, keys):
    for key in keys:
        levels = key.split('.')
        current_dict = api_capabilities_dict
        key_present = True
        for level in levels:
            if level not in current_dict:
                print(f"Key '{level}' not found in dictionary.")
                key_present = False
                break
            if 'properties' in current_dict[level]:
                current_dict = current_dict[level]['properties']
            else:
                current_dict = current_dict[level]
    return key_present


def extract_nested_fields(records, fields, keywords=None):
    """
    Recursively extracts nested fields from a list of dictionaries.

    :param records: List of dictionaries from which to extract fields
    :param fields: List of nested fields to extract, each specified with dot notation (e.g., 'openfda.brand_name')

    :return: List of dictionaries containing only the specified fields
    """
    extracted_records = []
    for record in records:
        extracted_record = {}
        for field in fields:
            keys = field.split('.')
            # print("keys", keys)
            value = record
            try:
                for key in keys:
                    value = value[key]
                if key != 'openfda' and key !='generic_name' and key !='brand_name':
                    if len(keywords)>0:
                        # print("key words:", keywords)
                        # print(value)
                        # print(type(value))
                        value = extract_sentences_with_keywords(value, keywords)
                extracted_record[field] = value
            except KeyError:
                extracted_record[field] = None
        if any(extracted_record.values()):
            extracted_records.append(extracted_record)
    return extracted_records

def map_properties_to_openfda_fields(arguments, search_fields):
    """
    Maps the provided arguments to the corresponding openFDA fields based on the search_fields mapping.

    :param arguments: The input arguments containing property names and values.
    :param search_fields: The mapping of property names to openFDA fields.

    :return: A dictionary with openFDA fields and corresponding values.
    """
    mapped_arguments = {}

    for key, value in list(arguments.items()):
        if key in search_fields:
            # print("key in search_fields:", key)
            openfda_fields = search_fields[key]
            if isinstance(openfda_fields, list):
                for field in openfda_fields:
                    mapped_arguments[field] = value
            else:
                mapped_arguments[openfda_fields] = value
            del arguments[key]
    arguments['search_fields'] = mapped_arguments
    return arguments

def extract_sentences_with_keywords(text_list, keywords):
    """
    Extracts sentences containing any of the specified keywords from the text.
    
    Parameters:
    - text (str): The input text from which to extract sentences.
    - keywords (list): A list of keywords to search for in the text.
    
    Returns:
    - list: A list of sentences containing any of the keywords.
    """
    sentences_with_keywords = []
    for text in text_list:
        # Compile a regular expression pattern for sentence splitting
        sentence_pattern = re.compile(r'(?<=[.!?]) +')
        # Split the text into sentences
        sentences = sentence_pattern.split(text)
        # Initialize a list to hold sentences with keywords
        
        
        # Iterate through each sentence
        for sentence in sentences:
            # Check if any of the keywords are present in the sentence
            if any(keyword.lower() in sentence.lower() for keyword in keywords):
                # If a keyword is found, add the sentence to the list
                sentences_with_keywords.append(sentence)
    
    return "......".join(sentences_with_keywords)


def search_openfda(params=None, endpoint_url=None, api_key=None, sort=None, limit=5, 
                   skip=None, count=None, exists=None, return_fields=None, 
                   exist_option='OR', search_keyword_option='AND', keywords_filter=True):
    # Initialize params if not provided
    if params is None:
        params = {}
    
    if return_fields=='ALL':
        exists = None

    # Initialize search fields and construct search query
    search_fields = params.get('search_fields', {})
    search_query = []
    keywords_list = []
    if search_fields:
        for field, value in search_fields.items():
            # Merge multiple continuous black spaces into one and use one '+'
            if keywords_filter and field !='openfda.brand_name' and field !='openfda.generic_name':
                keywords_list.extend(value.split())
            if field == 'openfda.generic_name':
                value = value.upper() # all generic names are in uppercase
            value = value.replace(" and ", " ") # remove 'and' in the search query
            value = value.replace(" AND ", " ") # remove 'AND' in the search query
            value = ' '.join(value.split())
            if search_keyword_option=='AND':
                search_query.append(f'{field}:({value.replace(" ", "+AND+")})')
            elif search_keyword_option=='OR':
                search_query.append(f'{field}:({value.replace(" ", "+")})')
            else:
                print("Invalid search_keyword_option. Please use 'AND' or 'OR'.")
        del params['search_fields']
    if search_query:
        params['search'] = "+".join(search_query)
        params['search'] = '(' + params['search'] + ')'
    # Validate the presence of at least one of search, count, or sort
    if not (params.get('search') or params.get('count') or params.get('sort') or search_fields):
        return {"error": "You must provide at least one of 'search', 'count', or 'sort' parameters."}

    # Set additional query parameters
    params['limit'] = params.get('limit', limit)
    params['sort'] = params.get('sort', sort)
    params['skip'] = params.get('skip', skip)
    params['count'] = params.get('count', count)
    if exists is not None:
        if isinstance(exists, str):
            exists = [exists]
        if 'search' in params:
            if exist_option == 'AND':
                params['search'] += "+AND+(" + "+AND+".join(
                    [f"_exists_:{keyword}" for keyword in exists])+")"
            elif exist_option == 'OR':
                params['search'] += "+AND+(" + "+".join(
                    [f"_exists_:{keyword}" for keyword in exists])+")"
        else:
            if exist_option == 'AND':
                params['search'] = "+AND+".join(
                    [f"_exists_:{keyword}" for keyword in exists])
            elif exist_option == 'OR':
                params['search'] = "+".join(
                    [f"_exists_:{keyword}" for keyword in exists])
        # Ensure that at least one of the search fields exists
        params['search'] += "+AND+(" + "+".join(
            [f"_exists_:{field}" for field in search_fields.keys()]) +")"
        # params['search']+="+AND+_exists_:openfda"

    # Construct full query with additional parameters
    query = "&".join(
        [f"{key}={value}" for key, value in params.items() if value is not None])
    full_url = f"{endpoint_url}?{query}"
    if api_key:
        full_url += f"&api_key={api_key}"

    print(full_url)

    response = requests.get(full_url)

    # Get the JSON response
    response_data = response.json()
    if 'error' in response_data:
        print("Invalid Query: ", response_data['error'])
        return None

    # Extract meta information
    meta_info = response_data.get('meta', {})
    meta_info = meta_info.get('results', {})

    # Extract results and return only the specified return fields
    results = response_data.get('results', [])
    if return_fields == 'ALL':
        return {
            'meta': meta_info,
            'results': results
        }
    required_fields = list(search_fields.keys()) + return_fields
    extracted_results = extract_nested_fields(
        results, required_fields, keywords_list)
    return {
        'meta': meta_info,
        'results': extracted_results
    }


class FDATool(BaseTool):
    def __init__(self, tool_config, endpoint_url, api_key=None):
        super().__init__(tool_config)
        fields = tool_config['fields']
        self.search_fields = fields.get('search_fields', {})
        self.return_fields = fields.get('return_fields', [])
        self.exists = fields.get('exists', None)
        if self.exists is None:
            self.exists = self.return_fields
        self.endpoint_url = endpoint_url
        self.api_key = api_key or os.getenv('FDA_API_KEY')

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        mapped_arguments = map_properties_to_openfda_fields(
            arguments, self.search_fields)
        return search_openfda(mapped_arguments,
                              endpoint_url=self.endpoint_url,
                              api_key=self.api_key,
                              exists=self.exists,
                              return_fields=self.return_fields, exist_option='OR')


class FDADrugLabelTool(FDATool):
    def __init__(self, tool_config, api_key=None):
        endpoint_url = 'https://api.fda.gov/drug/label.json'
        super().__init__(tool_config, endpoint_url, api_key)


class FDADrugLabelSearchTool(FDATool):
    def __init__(self, tool_config=None, api_key=None):
        self.tool_config = {
            "name": "FDADrugLabelSearch",
            "description": "Retrieve information of a specific drug.",
            "label": ["search", "drug"],
            "type": "FDADrugLabelSearch",
            "parameter": {
                "type": "object",
                "properties": {
                    "drug_name": {
                        "type": "string",
                        "description": "The name of the drug.",
                        "required": True
                    },
                    "return_fields": {
                        "type": "array",
                        "items": {
                            "type": "string",
                            "enum": ['ALL', 'abuse', 'accessories', 'active_ingredient', 'adverse_reactions', 'alarms', 'animal_pharmacology_and_or_toxicology', 'ask_doctor', 'ask_doctor_or_pharmacist', 'assembly_or_installation_instructions', 'boxed_warning', 'calibration_instructions', 'carcinogenesis_and_mutagenesis_and_impairment_of_fertility', 'cleaning', 'clinical_pharmacology', 'clinical_studies', 'compatible_accessories', 'components', 'contraindications', 'controlled_substance', 'dependence', 'description', 'diagram_of_device', 'disposal_and_waste_handling', 'do_not_use', 'dosage_and_administration', 'dosage_forms_and_strengths', 'drug_abuse_and_dependence', 'drug_and_or_laboratory_test_interactions', 'drug_interactions', 'effective_time', 'environmental_warning', 'food_safety_warning', 'general_precautions', 'geriatric_use', 'guaranteed_analysis_of_feed', 'health_care_provider_letter', 'health_claim', 'how_supplied', 'id', 'inactive_ingredient', 'indications_and_usage', 'information_for_owners_or_caregivers', 'information_for_patients', 'instructions_for_use', 'intended_use_of_the_device', 'keep_out_of_reach_of_children', 'labor_and_delivery', 'laboratory_tests', 'mechanism_of_action', 'microbiology', 'nonclinical_toxicology', 'nonteratogenic_effects', 'nursing_mothers', 'openfda', 'other_safety_information', 'overdosage', 'package_label_principal_display_panel', 'patient_medication_information', 'pediatric_use', 'pharmacodynamics', 'pharmacogenomics', 'pharmacokinetics', 'precautions', 'pregnancy', 'pregnancy_or_breast_feeding', 'purpose', 'questions', 'recent_major_changes', 'references', 'residue_warning', 'risks', 'route', 'safe_handling_warning', 'set_id', 'spl_indexing_data_elements', 'spl_medguide', 'spl_patient_package_insert', 'spl_product_data_elements', 'spl_unclassified_section', 'statement_of_identity', 'stop_use', 'storage_and_handling', 'summary_of_safety_and_effectiveness', 'teratogenic_effects', 'troubleshooting', 'use_in_specific_populations', 'user_safety_warnings', 'version', 'warnings', 'warnings_and_cautions', 'when_using', 'meta'],
                            "description": "Searchable field."
                        },
                        "description": "Fields to search within drug labels.",
                        "required": True
                    },
                    "limit": {
                        "type": "integer",
                        "description": "The number of records to return.",
                        "required": False
                    },
                    "skip": {
                        "type": "integer",
                        "description": "The number of records to skip.",
                        "required": False
                    }
                }
            },
            "fields": {
                "search_fields": {
                    "drug_name": ["openfda.brand_name", "openfda.generic_name"]
                },
            }
        }
        endpoint_url = 'https://api.fda.gov/drug/label.json'
        super().__init__(self.tool_config, endpoint_url, api_key)

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        mapped_arguments = map_properties_to_openfda_fields(
            arguments, self.search_fields)
        return_fields = arguments['return_fields']
        del arguments['return_fields']
        return search_openfda(mapped_arguments,
                              endpoint_url=self.endpoint_url,
                              api_key=self.api_key,
                              return_fields=return_fields, exists=return_fields, exist_option='OR')
        
class FDADrugLabelSearchIDTool(FDATool):
    def __init__(self, tool_config=None, api_key=None):
        self.tool_config = {
            "name": "FDADrugLabelSearchALLTool",
            "description": "Retrieve any related information to the query.",
            "label": ["search", "drug"],
            "type": "FDADrugLabelSearch",
            "parameter": {
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "key words need to be searched.",
                        "required": True
                    },
                    "return_fields": {
                        "type": "array",
                        "items": {
                            "type": "string",
                            "enum": ['ALL', 'abuse', 'accessories', 'active_ingredient', 'adverse_reactions', 'alarms', 'animal_pharmacology_and_or_toxicology', 'ask_doctor', 'ask_doctor_or_pharmacist', 'assembly_or_installation_instructions', 'boxed_warning', 'calibration_instructions', 'carcinogenesis_and_mutagenesis_and_impairment_of_fertility', 'cleaning', 'clinical_pharmacology', 'clinical_studies', 'compatible_accessories', 'components', 'contraindications', 'controlled_substance', 'dependence', 'description', 'diagram_of_device', 'disposal_and_waste_handling', 'do_not_use', 'dosage_and_administration', 'dosage_forms_and_strengths', 'drug_abuse_and_dependence', 'drug_and_or_laboratory_test_interactions', 'drug_interactions', 'effective_time', 'environmental_warning', 'food_safety_warning', 'general_precautions', 'geriatric_use', 'guaranteed_analysis_of_feed', 'health_care_provider_letter', 'health_claim', 'how_supplied', 'id', 'inactive_ingredient', 'indications_and_usage', 'information_for_owners_or_caregivers', 'information_for_patients', 'instructions_for_use', 'intended_use_of_the_device', 'keep_out_of_reach_of_children', 'labor_and_delivery', 'laboratory_tests', 'mechanism_of_action', 'microbiology', 'nonclinical_toxicology', 'nonteratogenic_effects', 'nursing_mothers', 'openfda', 'other_safety_information', 'overdosage', 'package_label_principal_display_panel', 'patient_medication_information', 'pediatric_use', 'pharmacodynamics', 'pharmacogenomics', 'pharmacokinetics', 'precautions', 'pregnancy', 'pregnancy_or_breast_feeding', 'purpose', 'questions', 'recent_major_changes', 'references', 'residue_warning', 'risks', 'route', 'safe_handling_warning', 'set_id', 'spl_indexing_data_elements', 'spl_medguide', 'spl_patient_package_insert', 'spl_product_data_elements', 'spl_unclassified_section', 'statement_of_identity', 'stop_use', 'storage_and_handling', 'summary_of_safety_and_effectiveness', 'teratogenic_effects', 'troubleshooting', 'use_in_specific_populations', 'user_safety_warnings', 'version', 'warnings', 'warnings_and_cautions', 'when_using', 'meta'],
                            "description": "Searchable field."
                        },
                        "description": "Fields to search within drug labels.",
                        "required": True
                    },
                    "limit": {
                        "type": "integer",
                        "description": "The number of records to return.",
                        "required": False
                    },
                    "skip": {
                        "type": "integer",
                        "description": "The number of records to skip.",
                        "required": False
                    }
                }
            },
            "fields": {
                "search_fields": {
                    "query": ['id']
                },
            }
        }
        endpoint_url = 'https://api.fda.gov/drug/label.json'
        super().__init__(self.tool_config, endpoint_url, api_key)

    def run(self, arguments):
        arguments = copy.deepcopy(arguments)
        mapped_arguments = map_properties_to_openfda_fields(
            arguments, self.search_fields)
        return_fields = arguments['return_fields']
        del arguments['return_fields']
        return search_openfda(mapped_arguments,
                              endpoint_url=self.endpoint_url,
                              api_key=self.api_key,
                              return_fields=return_fields, exists=return_fields, exist_option='OR')


class FDADrugLabelGetDrugGenericNameTool(FDADrugLabelTool):
    def __init__(self, tool_config=None, api_key=None):
        
        if tool_config is None:
            tool_config = {
                "name": "get_drug_generic_name",
                "description": "Get the drug’s generic name based on the drug's generic or brand name.",
                "parameter": {
                    "type": "object",
                    "properties": {
                    "drug_name": {
                        "type": "string",
                        "description": "The generic or brand name of the drug.",
                        "required": True
                    }
                    }
                },
                "fields": {
                    "search_fields": {
                    "drug_name": [
                        "openfda.brand_name",
                        "openfda.generic_name"
                    ]
                    },
                    "return_fields": [
                    "openfda.generic_name"
                    ]
                },
                "type": "FDADrugLabelGetDrugGenericNameTool",
                "label": [
                    "FDADrugLabel",
                    "purpose",
                    "FDA"
                ]
                }
        

        from .data.fda_drugs_with_brand_generic_names_for_tool import drug_list
        
        self.brand_to_generic = {drug['brand_name']: drug['generic_name'] for drug in drug_list}
        self.generic_to_brand = {drug['generic_name']: drug['brand_name'] for drug in drug_list}
        
        super().__init__(tool_config, api_key)

    def run(self, arguments):
        
        drug_info = {}
        
        drug_name = arguments.get('drug_name')
        if '-' in drug_name:
            drug_name = drug_name.split('-')[0] # to handle some drug names such as tarlatamab-dlle
        if drug_name in self.brand_to_generic:
            drug_info['openfda.generic_name'] = self.brand_to_generic[drug_name]
            drug_info['openfda.brand_name'] = drug_name
        elif drug_name in self.generic_to_brand:
            drug_info['openfda.brand_name'] = self.generic_to_brand[drug_name]
            drug_info['openfda.generic_name'] = drug_name
        else:
            results = super().run(arguments)
            if results is not None:
                drug_info['openfda.generic_name'] = results['results'][0]['openfda.generic_name'][0]
                drug_info['openfda.brand_name'] = results['results'][0]['openfda.brand_name'][0]
                print("drug_info", drug_info)
            else:
                drug_info = None
        return drug_info  